require 'spec_helper'
require 'puppet_spec/files'

require_relative('puppet_pal_2pec') unless RUBY_VERSION == '1.9.3'

#describe 'Puppet Pal' do
#  pending("Puppet::Pal is not available on Ruby 1.9.3") if RUBY_VERSION == '1.9.3'
##  before { skip("Puppet::Pal is not available on Ruby 1.9.3") if RUBY_VERSION == '1.9.3' }
#  require_relative 'puppet_pal_2pec'
#end
